@@ -17,8 +17,10 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <limits.h>
 
 #include "commands.h"
+#include "utils.h"
 
 static char argv0_buf[ARGV0_BUF_SIZE];
 
